"""Top-level package for DyLoPro.
DyLoPro is a generic visual analytics framework that 
facilitates the comprehensive exploration of the dynamics 
in event logs over time. It caters to all kinds of event logs, 
and provides the means to explore the dynamics from all 
different perspectives, both individually as well
as combined with the performance perspective. 
Consequently, also a multi-dimensional view on the 
dynamics in process performance can be adopted. The 
purpose of DyLoPro is to efficiently identify patterns, 
temporary or permanent changes and trends of interest
from different perspectives, which upon identification 
could be further analyzed in a manner deemed appropriate 
by the user.
"""
__author__ = """Brecht Wuyts """
__email__ = 'brecht.wuyts@kuleuven.be'
__version__ = '0.1.1'

from DyLoPro.DynamicLogPlotting import DynamicLogPlots